# Copyright (c) 2021, NakaMetPy Develoers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# Original source lisence: 
# Copyright (c) 2008,2015,2016,2018 MetPy Developers.
#

# kinematics
g = 9.81 # 重力加速度 m/s**2
g_acceralation = g # 重力加速度 m/s**2
Re = 6731.229 * 1000 # m
P0 = 100000 # Pa
PI = 3.141592653589793

# thermodynamics
sat_pressure_0c = 611.2 # units : Pa
R = 287 # J/K
Cp = 1004 # J/K
kappa = R / Cp
epsilone = 0.622 # (水：18/乾燥空気：28.8)
LatHeatC = 2.5*10**6 # J/kg
f0 = 1E-4
GammaD = g/Cp
Kelvin = 273.15


